(in-package :contextl)

(defun function-name-p (name)
  (or (symbolp name)
      (and (consp name)
           (eq (car name) 'setf)
           (symbolp (cadr name))
           (null (cddr name)))))

(defpackage "LAYERED-FUNCTION-DEFINERS" (:use))

(defparameter *layered-function-definers*
  (find-package "LAYERED-FUNCTION-DEFINERS"))

(defun get-layered-function-definer-name (name)
  (cond ((symbolp name)
         (intern (format nil "~A::~A"
                         (package-name (symbol-package name))
                         (symbol-name name))
                 *layered-function-definers*))
        ((consp name)
         (unless (function-name-p name)
           (error "Illegal function name: ~S." name))
         `(setf ,(get-layered-function-definer-name (cadr name))))
        (t (error "Illegal function name: ~S." name))))

(defmacro define-layered-function (name (&rest args) &body options)
  (let ((definer (get-layered-function-definer-name name)))
    (with-unique-names (layer-arg rest-arg)
      `(progn
         (defgeneric ,definer (,layer-arg ,@args)
           (:argument-precedence-order 
            ,@(let ((argument-precedence-order (assoc :argument-precedence-order options)))
                (if argument-precedence-order
                    (cdr argument-precedence-order)
                  (loop for arg in args
                        until (member arg lambda-list-keywords)
                        collect arg)))
            ,layer-arg)
           ,@(loop for option in options
                   unless (eq (car option) :argument-precedence-order)
                   collect option))
         (declaim (inline ,name))
         (defun ,name (&rest ,rest-arg)
           (declare (dynamic-extent ,rest-arg)
                    (optimize (speed 3) (debug 0) (safety 0)
                              (compilation-speed 0)))
           (apply #',definer (layer-context-prototype *active-context*) ,rest-arg))
         #',definer))))

(defmacro define-layered-method (&whole form name &body body)
  (loop with in-layerp = (eq (car body) :in-layer)
        with layer = (if in-layerp (cadr body) 't)
        with layer-arg = (gensym "LAYER-ARG-")
        for tail on (if in-layerp (cddr body) body)
        until (listp (car tail))
        collect (car tail) into qualifiers
        finally
        (when (member :in-layer qualifiers)
          (error "Incorrect occurrence of :in-layer in ~S. Must occur before qualifiers." form))
        (return `(defmethod ,(get-layered-function-definer-name name) ,@qualifiers
                   ((,layer-arg ,layer) ,@(car tail)) ,@(cdr tail)))))

(defun ensure-layered-function
       (name
        &rest initargs
        &key (lambda-list () lambda-list-p)
        (argument-precedence-order
         (loop for arg in lambda-list
               until (member arg lambda-list-keywords)
               collect arg))
        &allow-other-keys)
  (declare (dynamic-extent initargs))
  (unless lambda-list-p
    (error "The layered function ~S must be initialized with a lambda list." name))
  (let ((gf (let ((layer-arg (gensym "LAYER-ARG-")))
              (apply #'ensure-generic-function
                     (get-layered-function-definer-name name)
                     :argument-precedence-order
                     `(,@argument-precedence-order ,layer-arg)
                     :lambda-list
                     `(,layer-arg ,@lambda-list)
                     initargs)))
        (rest-arg (gensym)))
    (setf (fdefinition name)
          (compile nil `(lambda (&rest ,rest-arg)
                          (declare (dynamic-extent ,rest-arg)
                                   (optimize (speed 3) (debug 0) (safety 0)
                                             (compilation-speed 0)))
                          (apply (the function ,gf)
                                 (layer-context-prototype *active-context*)
                                 ,rest-arg))))
    gf))

(defun ensure-layered-method
       (layered-function-name
        lambda-expression 
        &key
        #-(or allegro clisp mcl) 
        (method-class
         (generic-function-method-class
          (fdefinition (get-layered-function-definer-name layered-function-name))))
        (in-layer (find-layer 't))
        (qualifiers ())
        (lambda-list (cadr lambda-expression))
        (specializers (loop for arg in lambda-list
                            until (member arg lambda-list-keywords)
                            collect (find-class 't))))
  (let ((layer-arg (gensym "LAYER-ARG-")))
    (destructuring-bind
        (lambda (&rest args) &body body)
        lambda-expression
      (unless (eq lambda 'lambda)
        (error "Incorrect lambda expression: ~S." lambda-expression))
      (ensure-method (fdefinition (get-layered-function-definer-name layered-function-name))
                     `(lambda (,layer-arg ,@args) ,@body)
                     #-(or allegro clisp mcl) :method-class
                     #-(or allegro clisp mcl) method-class
                     :qualifiers qualifiers
                     :lambda-list `(,layer-arg ,@lambda-list)
                     :specializers (cons in-layer specializers)))))
